#include "precompile.h"

#include <float.h>

#include "android.ai.h"
#include "android.h"
#include "room.h"
#include "metamgr.h"

/*
  机器人状态:
  1：锁定追踪
  2: 巡逻
  3: 思考
 */
AndroidAI::~AndroidAI()
{
}

void AndroidAI::Update(int delta_time)
{
    state_elapsed_time_ += delta_time;
    if (HumMaster()->dead) {
        return;
    }
    if (HumMaster()->playing_skill) {
        HumMaster()->UpdateSkill();
    }
    if (a8::HasBitFlag(HumMaster()->status, HS_Assaulting)) {
        HumMaster()->_UpdateAssaultMove();
    }
    switch (state_) {
    case AS_thinking:
        {
            if (state_elapsed_time_ > 1500 + rand() % 3000) {
                int rnd = rand();
                if (rnd % 100 < 30) {
                    ChangeToState(AS_moving);
                } else if (rnd % 100 < 50) {
                    ChangeToState(AS_attack);
                }
            }
        }
        break;
    case AS_moving:
        {
            bool need_chg = true;
            if (state_elapsed_time_ < 1000 + rand() % 2000) {
                DoMove();
                need_chg = false;
            } else if (locked_enemy_ && !locked_enemy_->dead &&
                       HumMaster()->pos.Distance(locked_enemy_->pos) < MetaMgr::Instance()->robot_pursuit_range) {
                if (HumMaster()->pos.Distance(locked_enemy_->pos) < MetaMgr::Instance()->robot_attack_range) {
                    ChangeToState(AS_attack);
                } else {
                    HumMaster()->move_dir = locked_enemy_->pos - HumMaster()->pos;
                    HumMaster()->move_dir.Normalize();
                    DoMove();
                }
                need_chg = false;
            }
            if (need_chg) {
                int rnd = rand();
                if (rnd % 100 < 30) {
                    ChangeToState(AS_thinking);
                } else if (rnd % 100 < 50) {
                    ChangeToState(AS_attack);
                }
            }
        }
        break;
    case AS_attack:
        {
            bool need_chg = true;
            if (state_elapsed_time_ < 1000) {
                DoAttack();
                need_chg = false;
            } else if (locked_enemy_ && locked_enemy_->dead &&
                       HumMaster()->pos.Distance(locked_enemy_->pos) < MetaMgr::Instance()->robot_pursuit_range) {
                if (HumMaster()->pos.Distance(locked_enemy_->pos) > MetaMgr::Instance()->robot_attack_range) {
                    ChangeToState(AS_moving);
                } else {
                    DoAttack();
                }
                need_chg = false;
            }
            if (need_chg) {
                int rnd = rand();
                if (rnd % 100 < 30) {
                    ChangeToState(AS_moving);
                } else if (rnd % 100 < 50) {
                    ChangeToState(AS_thinking);
                }
            }
        }
        break;
    default:
        break;
    }
}

void AndroidAI::LockEnemy(Human* enemy)
{
    locked_enemy_ = enemy;
    if (state_ != AS_attack) {
        ChangeToState(AS_attack);
    }
}

void AndroidAI::LockEnemy(int target_id)
{
    Human* enemy = HumMaster()->room->GetHumanByUniId(target_id);
    if (enemy) {
        LockEnemy(enemy);
        locked_frameno_ = HumMaster()->room->frameno;
    }
}

void AndroidAI::UnLockEnemy()
{
    locked_enemy_ = nullptr;
}

Human* AndroidAI::HumMaster()
{
    return (Human*)owner;
}

void AndroidAI::ChangeToState(AndroidState_e to_state)
{
    state_ = to_state;
    state_elapsed_time_ = 0;
    switch (state_) {
    case AS_moving:
        {
            if (locked_enemy_) {
                a8::Vec2 move_dir = locked_enemy_->pos - HumMaster()->pos;
                if (std::abs(move_dir.x) > 0.00001f ||
                    std::abs(move_dir.y) > 0.00001f) {
                    HumMaster()->move_dir = move_dir;
                    HumMaster()->move_dir.Normalize();
                } else {
                    return;
                }
            } else {
                HumMaster()->move_dir = a8::Vec2(1.0f, 0);
                HumMaster()->move_dir.Rotate(a8::RandAngle());
                HumMaster()->move_dir.Normalize();
            }
            HumMaster()->attack_dir = HumMaster()->move_dir;
        }
        break;
    default:
        break;
    }
}

void AndroidAI::DoMove()
{
    if (HumMaster()->room->gas_data.gas_mode == kGasInactive) {
        return;
    }
    if (HumMaster()->HasBuffEffect(kBET_Vertigo) || HumMaster()->HasBuffEffect(kBET_Dcgr)) {
        return;
    }
    if (owner->updated_times % 2 == 0 && !a8::HasBitFlag(HumMaster()->status, HS_Assaulting)) {
        int speed = std::max(1, (int)HumMaster()->GetSpeed());
        for (int i = 0; i < speed; ++i) {
            a8::Vec2 old_pos = HumMaster()->pos;
            HumMaster()->pos = HumMaster()->pos + HumMaster()->move_dir;
            if (HumMaster()->IsCollisionInMapService()) {
                HumMaster()->pos = old_pos;
                if (i == 0) {
                    HumMaster()->FindPath();
                }
                break;
            }
            HumMaster()->room->grid_service.MoveHuman(HumMaster());
        }
        HumMaster()->CheckSpecObject();
    }
}

void AndroidAI::DoAttack()
{
    if (HumMaster()->room->gas_data.gas_mode == kGasInactive) {
        return;
    }
    if (HumMaster()->HasBuffEffect(kBET_Vertigo) || HumMaster()->HasBuffEffect(kBET_Dcgr)) {
        return;
    }
    if (owner->updated_times % 10 == 0) {
        Human* enemy = locked_enemy_ ? locked_enemy_ : owner->room->FindEnemy(HumMaster());
        if (enemy &&
            !enemy->HasBuffEffect(kBET_Invincible) &&
            !enemy->HasBuffEffect(kBET_Hide) &&
            !enemy->HasBuffEffect(kBET_InGrass)
            ) {
            if (HumMaster()->CanUseSkill() &&
                HumMaster()->pos.Distance(enemy->pos) < HumMaster()->tank_meta_->i->atk_range()) {
                UseSkill(enemy);
            } else {
                Shot(enemy);
            }
        }
    }
}

void AndroidAI::Shot(Human* enemy)
{
    if (enemy->HasBuffEffect(kBET_Hide) ||
        enemy->HasBuffEffect(kBET_Invincible) ||
        enemy->dead
        ) {
        return;
    }
    if (((HumMaster()->room->frameno - HumMaster()->last_shot_frameno_) * (1000 / kSERVER_FRAME_RATE)) >=
        HumMaster()->GetFireRate()
        ) {
        a8::Vec2 shot_dir = enemy->pos - HumMaster()->pos;
        if (std::abs(shot_dir.x) > FLT_EPSILON ||
            std::abs(shot_dir.y) > FLT_EPSILON) {
            shot_dir.Normalize();
            shot_dir.Rotate((rand() % 10) / 180.0f);
            HumMaster()->attack_dir = shot_dir;
            if (HumMaster()->curr_weapon->meta->NeedTrace()) {
                HumMaster()->shot_target_id = enemy->entity_uniid;
            } else {
                HumMaster()->shot_target_id = 0;
            }
            HumMaster()->Shot();
        }
    }
}

void AndroidAI::UseSkill(Human* enemy)
{
    if (!HumMaster()->CurrentSkillMeta()) {
        return;
    }
    if (enemy->HasBuffEffect(kBET_Hide) ||
        enemy->HasBuffEffect(kBET_Invincible) ||
        enemy->dead) {
        return;
    }

    if (HumMaster()->CurrentSkillMeta()->i->skill_target() == kST_Self) {
        HumMaster()->curr_skill_phase = 0;
        HumMaster()->skill_target_id = HumMaster()->entity_uniid;
        HumMaster()->skill_dir = a8::Vec2();
        HumMaster()->skill_target_pos = HumMaster()->pos;
        HumMaster()->skill_param1 = 0;
        HumMaster()->playing_skill = false;
        HumMaster()->DoSkill();
    } else {
        a8::Vec2 shot_dir = enemy->pos - HumMaster()->pos;
        if (std::abs(shot_dir.x) > FLT_EPSILON ||
            std::abs(shot_dir.y) > FLT_EPSILON) {
            HumMaster()->curr_skill_phase = 0;
            HumMaster()->skill_target_id = enemy->entity_uniid;
            HumMaster()->skill_dir = shot_dir;
            HumMaster()->skill_target_pos = enemy->pos;
            HumMaster()->skill_param1 = 0;
            HumMaster()->playing_skill = false;
            HumMaster()->DoSkill();
        }
    }
}
