#include "precompile.h"

#include "frameevent.h"
#include "human.h"
#include "bullet.h"
#include "metamgr.h"
#include "typeconvert.h"
#include "room.h"

void FrameEvent::AddAirDrop(int appear_time, std::vector<a8::Vec2>& pos_list)
{
    cs::MFAirDrop* airdrop = airdrops_.Add();
    airdrop->set_appear_time(appear_time);
    for (auto& pos : pos_list) {
        TypeConvert::ToPb(pos, airdrop->add_pos_list());
    }
}

void FrameEvent::AddEmote(Human* hum, int emote_id)
{
    {
        auto& tuple = a8::FastAppend(emotes_);
        std::get<0>(tuple) = hum;
        auto& p = std::get<1>(tuple);

        p.set_emote_id(emote_id);
        p.set_player_id(hum->entity_uniid);
    }
    {
        int emote_idx = emotes_.size() - 1;
        for (auto& cell : hum->grid_list) {
            for (auto& hum : cell->human_list) {
                hum->emotes_.push_back(emote_idx);
            }
        }
    }
}

void FrameEvent::AddShot(Human* hum)
{
    {
        auto& tuple = a8::FastAppend(shots_);
        std::get<0>(tuple) = hum;
        auto& p = std::get<1>(tuple);

        p.set_player_id(hum->entity_uniid);
        hum->curr_weapon->ToPB(p.mutable_weapon());
        p.set_offhand(true);
        p.set_bullskin(10001);
    }
    {
        int shot_idx = shots_.size() - 1;
        for (auto& cell : hum->grid_list) {
            for (auto& hum : cell->human_list) {
                hum->shots_.push_back(shot_idx);
            }
        }
    }
}

void FrameEvent::AddBullet(Human* hum, int bullet_id, int bullet_lv,
                           a8::Vec2 born_pos, a8::Vec2 dir, float fly_distance,
                           int hit_time, int target_id, int skill_id)
{
    {
        auto& tuple = a8::FastAppend(bullets_);
        std::get<0>(tuple) = hum;
        auto& p = std::get<1>(tuple);

        p.set_player_id(hum->entity_uniid);
        #if 0
        p.set_bullet_id(hum->curr_weapon->meta->i->use_bullet());
        #endif
        TypeConvert::ToPb(born_pos, p.mutable_pos());
        TypeConvert::ToPb(dir, p.mutable_dir());
        p.set_bulletskin(10001);
        p.set_gun_id(bullet_id);
        p.set_gun_lv(bullet_lv);
        p.set_fly_distance(fly_distance);
        if (hit_time != 0) {
            p.set_hit_time(hit_time);
        }
        if (target_id != 0) {
            p.set_target_id(target_id);
        }
        if (skill_id != 0) {
            p.set_skill_id(skill_id);
        }
    }
    {
        int bullet_idx = bullets_.size() - 1;
        for (auto& cell : hum->grid_list) {
            for (auto& hum : cell->human_list) {
                hum->bullets_.push_back(bullet_idx);
            }
        }
    }
}

void FrameEvent::AddExplosion(int item_id, a8::Vec2 bomb_pos, int effect)
{
    {
        auto& p = a8::FastAppend(explosions_);

        p.set_item_id(item_id);
        TypeConvert::ToPb(bomb_pos, p.mutable_pos());
        if (effect != 0) {
            p.set_effect(effect);
        }
    }
    {
        int explosion_idx = explosions_.size() - 1;
        std::set<GridCell*> grid_list;
        room->grid_service.GetAllCellsByXy(bomb_pos.x, bomb_pos.y, grid_list);
        for (auto& cell : grid_list) {
            for (auto& hum : cell->human_list) {
                hum->explosions_.push_back(explosion_idx);
            }
        }
    }
}

void FrameEvent::AddSmoke(Bullet* bullet, int item_id, a8::Vec2 pos)
{
    {
        auto& tuple = a8::FastAppend(smokes_);
        std::get<0>(tuple) = bullet->master;
        auto& p = std::get<1>(tuple);

        p.set_item_id(item_id);
        TypeConvert::ToPb(pos, p.mutable_pos());
        p.set_player_id(bullet->master->entity_uniid);
    }
    {
        int idx = smokes_.size() - 1;
        for (auto& cell : bullet->master->grid_list) {
            for (auto& hum : cell->human_list) {
                hum->smokes_.push_back(idx);
            }
        }
    }
}

void FrameEvent::AddDead(Human* hum, int revive_time)
{
    {
        dead_objs_.push_back(
                             std::make_tuple(
                                             hum->entity_uniid,
                                             revive_time
                                             )
                             );
    }
    {
        int dead_idx = dead_objs_.size() - 1;
        for (auto& cell : hum->grid_list) {
            for (auto& hum : cell->human_list) {
                hum->dead_objs_.push_back(dead_idx);
            }
        }
    }
}

void FrameEvent::AddRevive(Human* hum)
{
    {
        revive_objs_.push_back(hum->entity_uniid);
    }
    {
        int revive_idx = revive_objs_.size() - 1;
        for (auto& cell : hum->grid_list) {
            for (auto& hum : cell->human_list) {
                hum->revive_objs_.push_back(revive_idx);
            }
        }
    }
}

void FrameEvent::AddBuff(Human* hum, Buff* buff)
{
    {
        cs::MFBuffChg chged_buff_pb;
        chged_buff_pb.set_obj_id(hum->entity_uniid);
        chged_buff_pb.set_chg(0);
        buff->FillMFBuff(chged_buff_pb.mutable_buff());
        chged_buffs_.push_back(std::make_tuple(hum, chged_buff_pb));
    }
    {
        int idx = chged_buffs_.size() - 1;
        for (auto& cell : hum->grid_list) {
            for (auto& hum : cell->human_list) {
                hum->chged_buffs_.push_back(idx);
            }
        }
    }
}

void FrameEvent::RemoveBuff(Human* hum, int buff_id)
{
    {
        cs::MFBuffChg chged_buff_pb;
        chged_buff_pb.set_obj_id(hum->entity_uniid);
        chged_buff_pb.set_chg(1);
        chged_buff_pb.mutable_buff()->set_buff_id(buff_id);
        chged_buffs_.push_back(std::make_tuple(hum, chged_buff_pb));
    }
    {
        int idx = chged_buffs_.size() - 1;
        for (auto& cell : hum->grid_list) {
            for (auto& hum : cell->human_list) {
                hum->chged_buffs_.push_back(idx);
            }
        }
    }
}

void FrameEvent::AddHpChg(Human* hum)
{
    chged_hps_.push_back(hum);
    int idx = chged_hps_.size() - 1;
    for (auto& cell : hum->grid_list) {
        for (auto& hum : cell->human_list) {
            hum->chged_hps_.push_back(idx);
        }
    }
}

void FrameEvent::AddInventoryChg(Human* hum)
{
    chged_inventorys_.push_back(hum);
    int idx = chged_inventorys_.size() - 1;
    hum->chged_inventorys_.push_back(idx);
}

void FrameEvent::AddSkillCdChg(Human* hum)
{
    chged_skillcds_.push_back(hum);
    int idx = chged_skillcds_.size() - 1;
    hum->chged_skillcds_.push_back(idx);
}

void FrameEvent::Clear()
{
    if (!explosions_.empty()) {
        explosions_.clear();
    }
    if (!smokes_.empty()) {
        smokes_.clear();
    }
    if (!emotes_.empty()) {
        emotes_.clear();
    }
    if (!bullets_.empty()) {
        bullets_.clear();
    }
    if (!shots_.empty()) {
        shots_.clear();
    }
    if (airdrops_.size() > 0) {
        airdrops_.Clear();
    }
    if (!revive_objs_.empty()) {
        revive_objs_.clear();
    }
    if (!dead_objs_.empty()) {
        dead_objs_.clear();
    }
    if (!chged_buffs_.empty()) {
        chged_buffs_.clear();
    }
    if (!chged_hps_.empty()) {
        chged_hps_.clear();
    }
    if (!chged_inventorys_.empty()) {
        chged_inventorys_.clear();
    }
    if (!chged_skillcds_.empty()) {
        chged_skillcds_.clear();
    }
}
