#include "precompile.h"

#include "framemaker.h"
#include "human.h"
#include "room.h"
#include "typeconvert.h"
#include "metamgr.h"

cs::SMUpdate* FrameMaker::MakeUpdateMsg(const Human* hum)
{
    cs::SMUpdate* msg = new cs::SMUpdate;
    {
        Room* room = hum->room;
        for (auto& itr : hum->new_objects_) {
            itr->FillMFObjectFull(msg->add_full_objects());
        }
        for (auto& itr : hum->part_objects_) {
            itr->FillMFObjectPart(msg->add_part_objects());
        }
        for (auto& itr : hum->del_objects_) {
            msg->add_del_objids(itr);
        }
        for (auto& itr : hum->out_objects_) {
            msg->add_out_objids(itr);
        }
        for (size_t idx : hum->shots_) {
            if (idx < room->frame_event.shots_.size()) {
                auto& tuple = room->frame_event.shots_[idx];
                if (hum->CanSee(std::get<0>(tuple))) {
                    *msg->add_shots() = std::get<1>(tuple);
                }
            }
        }
        for (size_t idx : hum->bullets_) {
            if (idx < room->frame_event.bullets_.size()) {
                auto& tuple = room->frame_event.bullets_[idx];
                if (hum->CanSee(std::get<0>(tuple))) {
                    *msg->add_bullets() = std::get<1>(tuple);
                }
            }
        }
        for (size_t idx : hum->explosions_) {
            if (idx < room->frame_event.explosions_.size()) {
                auto& p = room->frame_event.explosions_[idx];
                *msg->add_explosions() = p;
            }
        }
        for (size_t idx : hum->smokes_) {
            if (idx < room->frame_event.smokes_.size()) {
                auto& tuple = room->frame_event.smokes_[idx];
                if (hum->CanSee(std::get<0>(tuple))) {
                    *msg->add_smokes() = std::get<1>(tuple);
                }
            }
        }
        for (size_t idx : hum->emotes_) {
            if (idx < room->frame_event.emotes_.size()) {
                auto& tuple = room->frame_event.emotes_[idx];
                if (hum->CanSee(std::get<0>(tuple))) {
                    *msg->add_emotes() = std::get<1>(tuple);
                }
            }
        }
        for (size_t idx : hum->revive_objs_) {
            if (idx < room->frame_event.revive_objs_.size()) {
                msg->add_revive_objids(room->frame_event.revive_objs_[idx]);
            }
        }
        for (size_t idx : hum->dead_objs_) {
            if (idx < room->frame_event.dead_objs_.size()) {
                auto p = msg->add_dead_objs();
                p->add_values(std::get<0>(room->frame_event.dead_objs_[idx]));
                p->add_values(std::get<1>(room->frame_event.dead_objs_[idx]));
            }
        }
        for (size_t idx : hum->chged_buffs_) {
            if (idx < room->frame_event.chged_buffs_.size()) {
                auto p = msg->add_chged_buff_list();
                *p = std::get<1>(room->frame_event.chged_buffs_[idx]);
            }
        }
        for (size_t idx : hum->chged_hps_) {
            if (idx < room->frame_event.chged_hps_.size()) {
                Human* target = room->frame_event.chged_hps_[idx];
                {
                    auto p = msg->add_chged_property_list();
                    p->set_obj_id(target->entity_uniid);
                    p->set_property_type(kPropHp);
                    p->set_value(target->GetHP());
                }
                {
                    auto p = msg->add_chged_property_list();
                    p->set_obj_id(target->entity_uniid);
                    p->set_property_type(kPropMaxHp);
                    p->set_value(target->GetMaxHP());
                }
            }
        }
        for (size_t idx : hum->chged_inventorys_) {
            if (idx < room->frame_event.chged_inventorys_.size()) {
                Human* target = room->frame_event.chged_inventorys_[idx];
                if (target->curr_weapon) {
                    auto p = msg->add_chged_property_list();
                    p->set_obj_id(target->entity_uniid);
                    p->set_property_type(kPropInventory);
                    p->set_property_subtype(kWEAPON_SLOT);
                    p->set_value(target->curr_weapon->ammo);
                }
            }
        }
        for (size_t idx : hum->chged_skillcds_) {
            if (idx < room->frame_event.chged_skillcds_.size()) {
                Human* target = room->frame_event.chged_skillcds_[idx];
                {
                    auto p = msg->add_chged_property_list();
                    p->set_obj_id(target->entity_uniid);
                    p->set_property_type(kPropSkillLeftTime);
                    p->set_value(target->GetSkillLeftTime());
                }
                {
                    auto p = msg->add_chged_property_list();
                    p->set_obj_id(target->entity_uniid);
                    p->set_property_type(kPropSkillCd);
                    p->set_value(target->GetSkillCd());
                }
            }
        }
        if (room->frame_event.airdrops_.size() > 0) {
            *msg->mutable_airdrop() = room->frame_event.airdrops_.Get(0);
        }
        {
            if (room->BattleStarted()) {
                int game_left_time = MetaMgr::Instance()->game_duration * 1000 -
                    (room->frameno - room->battle_start_frameno_) * kFRAME_RATE_MS;
                game_left_time = std::max(0, game_left_time);
                msg->set_game_left_time(game_left_time);
            } else {
                msg->set_game_left_time(MetaMgr::Instance()->game_duration * 1000);
            }
        }
    }
    return msg;
}
