#include "precompile.h"

#include "metadata.h"
#include "metamgr.h"

namespace MetaData
{
    void Parameter::Init()
    {
        int_val = a8::XValue(i->param_value());
        float_val = a8::XValue(i->param_value());
        str_val = a8::XValue(i->param_value()).GetString();
    }

    void Map::Init()
    {
    }

    void Equip::Init()
    {
        {
            std::vector<std::string> strings0;
            a8::Split(i->bullet_born_offset(), strings0, '$');
            for (auto& str0 : strings0) {
                std::vector<std::string> strings;
                a8::Split(str0, strings, '|');
                auto& p = a8::FastAppend(bullet_born_offset);
                std::get<0>(p) = i->bullet_delay() * (bullet_born_offset.size() - 1);
                for (auto& str : strings) {
                    std::vector<std::string> strings2;
                    a8::Split(str, strings2, ':');
                    assert(strings2.size() >= 2);
                    auto& born_point = a8::FastAppend(std::get<1>(p));
                    born_point.born_offset = a8::Vec2(a8::XValue(strings2[0]).GetDouble(),
                                                      a8::XValue(strings2[1]).GetDouble()
                                                      );
                    born_point.angle = strings2.size() > 2 ? a8::XValue(strings2[2]).GetDouble() : 0;
                }
            }
        }
    }

    bool Equip::CanDrop()
    {
        return i->equip_type() == kEquipType_Buff && MetaMgr::Instance()->fighting_mode;
    }

    bool Equip::NeedTrace()
    {
        return i->equip_type() == kEquipType_Bullet &&
            (i->equip_subtype() == kBulletType_Missile ||
             i->equip_subtype() == kBulletType_Trace);
    }

    void Player::Init()
    {
        {
            std::vector<std::string> strings;
            a8::Split(i->volume(), strings, '|');
            for (auto& str : strings) {
                std::vector<std::string> strings2;
                a8::Split(str, strings2, ':');
                assert(strings2.size() == 2);
                if (strings2.size() >= 2) {
                    size_t slot = a8::XValue(strings2[0]);
                    size_t num = a8::XValue(strings2[1]);
                    if (slot >= 0 && slot < volume.size()){
                        volume[slot] = num;
                    }
                }
            }
        }
    }

    void Robot::Init()
    {

    }

    void Drop::Init()
    {
        std::vector<std::string> item_list;
        {
            a8::Split(i->item_id(), item_list, '|');
        }
        std::vector<std::string> num_list;
        {
            a8::Split(i->num(), num_list, '|');
        }
        std::vector<std::string> weight_list;
        {
            a8::Split(i->weight(), weight_list, '|');
        }
        assert(item_list.size() == num_list.size() &&
               item_list.size() == weight_list.size());
        total_weight = 0;
        for (size_t i = 0; i < item_list.size(); ++i) {
            total_weight += a8::XValue(weight_list[i]).GetInt();
            std::vector<int> itemids;
            {
                std::vector<std::string> strings;
                a8::Split(item_list[i], strings, ':');
                for (auto& str : strings) {
                    itemids.push_back(a8::XValue(str).GetInt());
                }
            }
            std::vector<int> nums;
            {
                std::vector<std::string> strings;
                a8::Split(num_list[i], strings, ':');
                for (auto& str : strings) {
                    nums.push_back(a8::XValue(str).GetInt());
                }
            }
            assert(itemids.size() == nums.size());
            std::vector<int> item_lvs;
            for (size_t ii = 0; ii < itemids.size(); ++ii) {
                item_lvs.push_back(1);
            }
            if (this->i->type() == 1) {
                auto item_tuple = std::make_tuple(
                                                  itemids,
                                                  nums,
                                                  item_lvs,
                                                  a8::XValue(weight_list[i]).GetInt()
                                                  );
                items.push_back(item_tuple);
            } else {
                auto item_tuple = std::make_tuple(
                                                  itemids,
                                                  nums,
                                                  item_lvs,
                                                  total_weight
                                                  );
                items.push_back(item_tuple);
            }
        }
        assert(i->type() == 1 || i->type() == 2);
    }

    void Drop::RandItems(std::vector<std::tuple<int, int, int>>& drop_items)
    {
        if (i->type() == 1) {
            for (auto& item : items) {
                if ((rand() % 10000) <= std::get<3>(item)) {
                    for (size_t i = 0; i < std::get<0>(item).size(); ++i) {
                        drop_items.push_back(std::make_tuple(
                                                             std::get<0>(item)[i],
                                                             std::get<1>(item)[i],
                                                             std::get<2>(item)[i]
                                                             ));
                    }
                }
            }
        } else if (total_weight > 0) {
            int rnd = rand() % total_weight;
            for (auto& item : items) {
                if (std::get<3>(item) >= rnd) {
                    for (size_t i = 0; i < std::get<0>(item).size(); ++i) {
                        drop_items.push_back(std::make_tuple(
                                                             std::get<0>(item)[i],
                                                             std::get<1>(item)[i],
                                                             std::get<2>(item)[i]
                                                             ));
                    }
                    break;
                }
            }
        }
    }

    void AirDrop::Init()
    {
        std::vector<std::string> strings;
        a8::Split(i->drop(), strings, '|');
        for (std::string& tmpstr : strings) {
            std::vector<std::string> strings2;
            a8::Split(tmpstr, strings2, ':');
            if (strings2.size() != 2) {
                abort();
            }
            drop_list.push_back(std::make_tuple(
                                                a8::XValue(strings2[0]).GetInt(),
                                                a8::XValue(strings2[1]).GetInt()
                                                )
                                );
        }
    }

    void AirDrop::RandItems(std::vector<std::tuple<int, int>>& drop_items)
    {
        for (auto& tuple : drop_list) {
            drop_items.push_back(
                                 std::make_tuple(
                                                 std::get<0>(tuple),
                                                 std::get<1>(tuple)
                                                 )
                                 );
        }
    }

    void Building::Init()
    {
        for (auto& door_meta : i->doorobj()) {
            Door* p_door = nullptr;
            for (auto& door : doors) {
                if (door.door_id == door_meta.id()) {
                    p_door = &door;
                    break;
                }
            }
            if (!p_door) {
                p_door = &a8::FastAppend(doors);
                p_door->door_id = door_meta.id();
            }
            switch (door_meta.type()) {
            case 1:
                {
                    p_door->state0 = &door_meta;
                }
                break;
            case 2:
                {
                    p_door->state1 = &door_meta;
                }
                break;
            }
        }//end for

        for (auto& door : doors) {
            if (door.door_id == 0 ||
                !door.state0 || !door.state1) {
                abort();
            } else {
                #if 0
                if (door.state1->x() < door.state0.x()) {
                    if (door.state1.y() < door.state0.y()) {

                    } else {

                    }
                }
                #endif
            }
        }

        metatable::BuildingJson* building_meta = (metatable::BuildingJson*)i;
        for (auto& loot : *building_meta->mutable_lootobj()) {
            loot.set__rand_space(0);
            std::vector<std::string> strings;
            a8::Split(loot.things(), strings, '\n');
            for (auto& str : strings) {
                std::vector<std::string> strings2;
                a8::Split(str, strings2, ':');
                assert(strings2.size() == 2);
                loot.set__rand_space(
                                     loot._rand_space() +
                                     a8::XValue(strings2[1]).GetInt()
                                     );
                auto p = loot.add__things();
                p->set_key(a8::XValue(strings2[0]));
                p->set_value(loot._rand_space());
            }
        }
    }

    void MapTplThing::Init()
    {
        {
            rand_space = 0;
            std::vector<std::string> strings;
            a8::Split(i->things(), strings, '\n');
            for (auto& str : strings) {
                std::vector<std::string> strings2;
                a8::Split(str, strings2, ':');
                assert(strings2.size() == 2);
                rand_space += a8::XValue(strings2[1]).GetInt();
                things.push_back(std::make_tuple(
                                                 a8::XValue(strings2[0]),
                                                 rand_space
                                                 )
                                 );
            }
        }
        {
            std::vector<std::string> strings;
            a8::Split(i->born_angle(), strings, '\n');
            for (auto& str : strings) {
                born_angle.push_back(a8::XValue(str).GetDouble());
            }
        }
    }

    int MapTplThing::RandThing()
    {
        if (rand_space <= 0) {
            return 0;
        }
        int rnd = rand() % rand_space;
        for (auto& tuple : things) {
            if (std::get<1>(tuple) > rnd) {
                return std::get<0>(tuple);
            }
        }
        return 0;
    }

    void Skill::Init()
    {
        {
            std::vector<std::string> strings;
            a8::Split(i->buff_list(), strings, '|');
            for (auto& str : strings) {
                buff_list.insert(a8::XValue(str));
            }
        }
        {
            const google::protobuf::Descriptor* descriptor = i->GetDescriptor();
            const google::protobuf::Reflection* reflection = i->GetReflection();
            for (int idx = 1; idx <= 100; ++idx) {
                MetaData::SkillPhase phase;
                {
                    auto field_desc = descriptor->FindFieldByName(a8::Format("phase%d_time_offset", {idx}));
                    if (!field_desc) {
                        break;
                    }
                    phase.phase_idx = phases.size() + 1;
                    phase.time_offset = reflection->GetInt32(*(metatable::Skill*)i, field_desc);
                }
                {
                    auto field_desc = descriptor->FindFieldByName(a8::Format("phase%d_func", {idx}));
                    phase.func_id = (SkillFunc_e)reflection->GetInt32(*(metatable::Skill*)i, field_desc);
                }
                {
                    auto field_desc = descriptor->FindFieldByName(a8::Format("phase%d_param1", {idx}));
                    phase.param1_str = reflection->GetString(*(metatable::Skill*)i, field_desc);
                }
                {
                    auto field_desc = descriptor->FindFieldByName(a8::Format("phase%d_param2", {idx}));
                    phase.param2_str = reflection->GetString(*(metatable::Skill*)i, field_desc);
                }
                {
                    auto field_desc = descriptor->FindFieldByName(a8::Format("phase%d_param3", {idx}));
                    phase.param3_str = reflection->GetString(*(metatable::Skill*)i, field_desc);
                }
                if (phase.func_id != kSkill_FuncNone) {
                    phase.param1 = a8::XValue(phase.param1_str);
                    phase.param2 = a8::XValue(phase.param2_str);
                    phase.param3 = a8::XValue(phase.param3_str);
                    phases.push_back(phase);
                }
            }
        }
        {
            for (int buff_id : buff_list) {
                MetaData::Buff* buff_meta = MetaMgr::Instance()->GetBuff(buff_id);
                if (!buff_meta) {
                    abort();
                }
                auto itr = trigger_type_buffs.find(buff_meta->i->trigger_type());
                if (itr != trigger_type_buffs.end()) {
                    itr->second.insert(buff_meta);
                } else {
                    trigger_type_buffs[buff_meta->i->trigger_type()] = std::set<MetaData::Buff*>({buff_meta});
                }
            }
        }
    }

    void Buff::Init()
    {
        param1 = a8::XValue(i->buff_param1()).GetDouble();
        param2 = a8::XValue(i->buff_param2()).GetDouble();
        param3 = a8::XValue(i->buff_param3()).GetDouble();
        param4 = a8::XValue(i->buff_param4()).GetDouble();
    }

    bool Buff::EffectCanStack()
    {
        return i->buff_effect() == kBET_ChgAttr;
    }

    void Driver::Init()
    {
        {
            std::vector<std::string> strings;
            a8::Split(i->attr_type(), strings, '|');
            for (auto& str : strings) {
                std::vector<std::string> strings2;
                a8::Split(str, strings2, ':');
                if (strings2.size() != 3) {
                    abort();
                }
                int attr_type = a8::XValue(strings2[0]);
                float base_value = a8::XValue(strings2[1]).GetDouble();
                float up_value = a8::XValue(strings2[2]).GetDouble();
                attr_up[attr_type] = std::make_tuple(base_value, up_value);
            }
        }
        {
            std::vector<std::string> strings;
            a8::Split(i->s_attr(), strings, '|');
            for (auto& str : strings) {
                std::vector<std::string> strings2;
                a8::Split(str, strings2, ':');
                if (strings2.size() != 3) {
                    abort();
                }
                int attr_type = a8::XValue(strings2[0]);
                float attr_value = a8::XValue(strings2[1]).GetDouble();
                int attr_level = a8::XValue(strings2[2]).GetDouble();
                passive_skill_attr[attr_type] = std::make_tuple(attr_value, attr_level);
            }
        }
    }

    void Tank::Init()
    {
        {
            std::vector<std::string> strings;
            a8::Split(i->attr_origin(), strings, '|');
            for (auto& str : strings) {
                std::vector<std::string> strings2;
                a8::Split(str, strings2, ':');
                if (strings2.size() != 2) {
                    abort();
                }
                int attr_type = a8::XValue(strings2[0]);
                float attr_value = a8::XValue(strings2[1]).GetDouble();
                attr_init[attr_type] = attr_value;
            }
        }
        {
            std::vector<std::string> strings;
            a8::Split(i->attr_up(), strings, '|');
            for (auto& str : strings) {
                std::vector<std::string> strings2;
                a8::Split(str, strings2, ':');
                if (strings2.size() != 3) {
                    abort();
                }
                int attr_type = a8::XValue(strings2[0]);
                int attr_level = a8::XValue(strings2[1]).GetDouble();
                float attr_value = a8::XValue(strings2[2]).GetDouble();
                attr_up[attr_type] = std::make_tuple(attr_value, attr_level);
            }
        }
    }

}
