#include "precompile.h"

#include <a8/stringlist.h>
#include <a8/csvreader.h>

#include "framework/cpp/utils.h"

#include "metamgr.h"
#include "app.h"

class MetaDataLoader
{
public:
    std::list<metatable::Parameter> parameter_meta_list;
    std::list<MetaData::Parameter> parameter_list;
    std::list<metatable::Map> map_meta_list;
    std::list<MetaData::Map> map_list;
    std::list<metatable::SafeArea> safearea_meta_list;
    std::list<MetaData::SafeArea> safearea_list;
    std::list<metatable::Item> item_meta_list;
    std::list<MetaData::Item> item_list;
    std::list<metatable::Equip> equip_meta_list;
    std::list<MetaData::Equip> equip_list;
    std::list<metatable::Player> player_meta_list;
    std::list<MetaData::Player> player_list;
    std::list<metatable::MapThing> mapthing_meta_list;
    std::list<MetaData::MapThing> mapthing_list;
    std::list<metatable::BuildingJson> building_meta_list;
    std::list<MetaData::Building> building_list;
    std::list<metatable::Drop> drop_meta_list;
    std::list<MetaData::Drop> drop_list;
    std::list<metatable::AirDrop> airdrop_meta_list;
    std::list<MetaData::AirDrop> airdrop_list;
    std::list<metatable::Tank> tank_meta_list;
    std::list<MetaData::Tank> tank_list;
    std::list<metatable::TankSkin> tanktank_meta_list;
    std::list<MetaData::TankSkin> tankskin_list;
    std::list<metatable::Cup> cup_meta_list;
    std::list<MetaData::Cup> cup_list;
    std::list<metatable::Driver> driver_meta_list;
    std::list<MetaData::Driver> driver_list;
    std::list<metatable::Skill> skill_meta_list;
    std::list<MetaData::Skill> skill_list;
    std::list<metatable::Buff> buff_meta_list;
    std::list<MetaData::Buff> buff_list;
    std::list<metatable::Attr> attr_meta_list;
    std::list<MetaData::Attr> attr_list;
    std::list<metatable::RankReward> rankreward_meta_list;
    std::list<MetaData::RankReward> rankreward_list;
    std::list<metatable::KillReward> killreward_meta_list;
    std::list<MetaData::KillReward> killreward_list;
    std::list<metatable::Robot> robot_meta_list;
    std::vector<MetaData::Robot> robot_list;

    std::map<std::string, MetaData::Parameter*> parameter_hash;
    std::map<int, MetaData::Map*> gamemap_hash;
    std::map<int, MetaData::SafeArea*> safearea_hash;
    std::map<int, MetaData::Item*> item_hash;
    std::map<int, MetaData::Equip*> equip_hash;
    std::map<int, MetaData::Equip*> equip_slot_hash;
    std::map<int, MetaData::Player*> player_hash;
    std::map<int, MetaData::MapThing*> mapthing_hash;
    std::map<int, MetaData::Building*> building_hash;
    std::map<int, MetaData::Drop*> drop_hash;
    std::map<int, std::vector<MetaData::AirDrop*>> airdrop_hash;
    std::map<std::string, std::list<metatable::MapTplThingJson>> maptpl_meta_hash;
    std::map<std::string, std::vector<MetaData::MapTplThing>> maptpl_hash;
    std::map<std::string, std::vector<MetaData::MapTplThing>> born_point_hash;
    std::map<std::string, std::map<int, MetaData::MapTplThing>> airdrop_point_hash;
    std::map<std::string, std::tuple<int, int>> maptpl_size_hash;
    std::map<int, MetaData::Tank*> tank_hash;
    std::map<int, MetaData::TankSkin*> tankskin_hash;
    std::map<int, MetaData::Cup*> cup_hash;
    std::vector<int> tankid_list;
    std::map<int, MetaData::Driver*> driver_hash;
    std::map<int, MetaData::Skill*> skill_hash;
    std::map<int, MetaData::Buff*> buff_hash;
    std::map<int, MetaData::Attr*> attr_id_hash;
    std::map<std::string, MetaData::Attr*> attr_name_hash;
    std::map<int, MetaData::RankReward*> rankreward_hash;
    std::map<int, MetaData::KillReward*> killreward_hash;
    std::map<int, MetaData::Robot*> robot_hash;

    void Load()
    {
        if (!f8::IsOnlineEnv()) {
            if (App::Instance()->HasFlag(2)) {
                res_path = a8::Format("/root/pub/%d/%d/conf_test/game%d/gameserver/res/",
                                     {
                                      GAME_ID,
                                      App::Instance()->instance_id,
                                      GAME_ID
                                     });
            } else {
                res_path = a8::Format("/var/data/conf_test/game%d/gameserver/res/", {GAME_ID});
            }
        } else {
            res_path = "../res/";
        }

        f8::ReadCsvMetaFile(res_path + "parameter@parameter.csv", parameter_meta_list);
        f8::ReadCsvMetaFile(res_path + "map@map.csv", map_meta_list);
        f8::ReadCsvMetaFile(res_path + "safearea@safearea.csv", safearea_meta_list);
        f8::ReadCsvMetaFile(res_path + "tank@tank.csv", tank_meta_list);
        f8::ReadCsvMetaFile(res_path + "tankskin@tankskin.csv", tanktank_meta_list);
        f8::ReadCsvMetaFile(res_path + "cup@cup.csv", cup_meta_list);
        f8::ReadCsvMetaFile(res_path + "driver@driver.csv", driver_meta_list);
        f8::ReadCsvMetaFile(res_path + "item@item.csv", item_meta_list);
        f8::ReadCsvMetaFile(res_path + "equip@equip.csv", equip_meta_list);
        f8::ReadCsvMetaFile(res_path + "player@player.csv", player_meta_list);
        f8::ReadCsvMetaFile(res_path + "player@player.csv", player_meta_list);
        f8::ReadCsvMetaFile(res_path + "mapThing@mapThing.csv", mapthing_meta_list);
        f8::ReadCsvMetaFile(res_path + "drop@drop.csv", drop_meta_list);
        f8::ReadCsvMetaFile(res_path + "airdrop@airdrop.csv", airdrop_meta_list);
        f8::ReadJsonMetaFile(res_path + "maps.json", building_meta_list);
        f8::ReadCsvMetaFile(res_path + "skill@skill.csv", skill_meta_list);
        f8::ReadCsvMetaFile(res_path + "buff@buff.csv", buff_meta_list);
        f8::ReadCsvMetaFile(res_path + "attr@attr.csv", attr_meta_list);
        f8::ReadCsvMetaFile(res_path + "rankReward@rankReward.csv", rankreward_meta_list);
        f8::ReadCsvMetaFile(res_path + "killReward@killReward.csv", killreward_meta_list);
        f8::ReadCsvMetaFile(res_path + "robot@robot.csv", robot_meta_list);
        BindToMetaData();
        #if 1
        {
            MetaMgr::Instance()->gas_inactive_time = MetaMgr::Instance()->GetSysParamAsInt("gas_inactive_time");
            MetaMgr::Instance()->K = MetaMgr::Instance()->GetSysParamAsFloat("K");
            MetaMgr::Instance()->kill_param = MetaMgr::Instance()->GetSysParamAsFloat("kill_parameter");
            MetaMgr::Instance()->rank_param = MetaMgr::Instance()->GetSysParamAsFloat("rank_parameter");
            MetaMgr::Instance()->fighting_mode = MetaMgr::Instance()->GetSysParamAsInt("fighting_mode", 1);
            MetaMgr::Instance()->revive_time = MetaMgr::Instance()->GetSysParamAsInt("revive_time", 5);
            MetaMgr::Instance()->game_duration = MetaMgr::Instance()->GetSysParamAsInt("revive_time", 60 * 2);
            MetaMgr::Instance()->grass_invisible_time = MetaMgr::Instance()->GetSysParamAsInt("grass_invisible_time", 0.5f);
            MetaMgr::Instance()->grass_show_time = MetaMgr::Instance()->GetSysParamAsInt("grass_show_time", 0.5f);
            MetaMgr::Instance()->grass_invisible_time2 = MetaMgr::Instance()->GetSysParamAsInt("grass_invisible_time2", 2.0f);
            MetaMgr::Instance()->robot_pursuit_range = MetaMgr::Instance()->
                GetSysParamAsInt("robot_pursuit_range",
                                 MetaMgr::Instance()->robot_pursuit_range
                                 );
            MetaMgr::Instance()->robot_attack_range = MetaMgr::Instance()->
                GetSysParamAsInt("robot_attack_range",
                                 MetaMgr::Instance()->robot_attack_range
                                 );
            if (MetaMgr::Instance()->K < 0.01f) {
                abort();
            }
        }
        #endif
    }

private:

    void BindToMetaData()
    {

        for (auto& meta : parameter_meta_list) {
            MetaData::Parameter& item = a8::FastAppend(parameter_list);
            item.i = &meta;
            item.Init();
            parameter_hash[item.i->param_name()] = &item;
        }

        for (auto& meta : map_meta_list) {
            MetaData::Map& item = a8::FastAppend(map_list);
            item.i = &meta;
            item.Init();
            gamemap_hash[item.i->map_id()] = &item;
            #if 1
            {
                {
                    auto itr = maptpl_meta_hash.find(meta.template_name());
                    if (itr == maptpl_meta_hash.end()) {
                        maptpl_meta_hash[meta.template_name()] = std::list<metatable::MapTplThingJson>();
                        itr = maptpl_meta_hash.find(meta.template_name());
                    } else {
                        itr->second.clear();
                    }
                    f8::ReadJsonMetaFile(res_path + meta.template_name() + ".json", itr->second);
                }
                for (auto& pair : maptpl_meta_hash) {
                    std::vector<MetaData::MapTplThing> things;
                    std::vector<MetaData::MapTplThing> born_point;
                    std::map<int, MetaData::MapTplThing> airdrop_point;
                    int map_width = 0;
                    int map_height = 0;
                    for (auto& itr : pair.second) {
                        if (itr.object_type() == "airdrop") {
                            MetaData::MapTplThing thing;
                            thing.i = &itr;
                            thing.Init();
                            airdrop_point[itr.object_uniid()] = thing;
                        } else {
                            if (itr.is_born_point()) {
                                auto& thing = a8::FastAppend(born_point);
                                thing.i = &itr;
                                thing.Init();
                            } else if (itr.is_map_info()) {
                                map_width = itr.map_width();
                                map_height = itr.map_height();
                            } else {
                                auto& thing = a8::FastAppend(things);
                                thing.i = &itr;
                                thing.Init();
                            }
                        }
                    }
                    if (map_width < kMAP_GRID_WIDTH || map_height < kMAP_GRID_WIDTH) {
                        abort();
                    }
                    maptpl_hash[pair.first] = things;
                    born_point_hash[pair.first] = born_point;
                    airdrop_point_hash[pair.first] = airdrop_point;
                    maptpl_size_hash[pair.first] = std::make_tuple(map_width, map_height);
                    if (born_point.size() != kROOM_MAX_PLAYER_NUM) {
                        abort();
                    }
                }
            }
            #endif
        }

        for (auto& meta : safearea_meta_list) {
            MetaData::SafeArea& item = a8::FastAppend(safearea_list);
            item.i = &meta;
            if (App::Instance()->HasFlag(7)) {
                meta.set_wait_time(3);
                meta.set_shrink_speed(500);
            }
            safearea_hash[item.i->id()] = &item;
        }

        for (auto& meta : tank_meta_list) {
            MetaData::Tank& item = a8::FastAppend(tank_list);
            item.i = &meta;
            item.Init();
            tank_hash[item.i->id()] = &item;
            tankid_list.push_back(meta.id());
        }

        for (auto& meta : tanktank_meta_list) {
            MetaData::TankSkin& item = a8::FastAppend(tankskin_list);
            item.i = &meta;
            tankskin_hash[item.i->tank_id()] = &item;
        }

        for (auto& meta : cup_meta_list) {
            MetaData::Cup& item = a8::FastAppend(cup_list);
            item.i = &meta;
            cup_hash[item.i->rank()] = &item;
        }

        for (auto& meta : driver_meta_list) {
            MetaData::Driver& item = a8::FastAppend(driver_list);
            item.i = &meta;
            item.Init();
            driver_hash[item.i->id()] = &item;
        }

        for (auto& meta : item_meta_list) {
            MetaData::Item& item = a8::FastAppend(item_list);
            item.i = &meta;
            item_hash[item.i->id()] = &item;
        }

        for (auto& meta : equip_meta_list) {
            #if 0
            if (meta.inventory_slot().empty()) {
                meta.set__inventory_slot(-1);
            } else {
                meta.set__inventory_slot(a8::XValue(meta.inventory_slot()));
            }
            #endif
            MetaData::Equip& item = a8::FastAppend(equip_list);
            item.i = &meta;
            item.Init();
            equip_hash[item.i->id()] = &item;
            #if 0
            if (meta._inventory_slot() > -1) {
                equip_slot_hash[meta._inventory_slot()] = &item;
            }
            #endif
        }

        for (auto& meta : player_meta_list) {
            MetaData::Player& item = a8::FastAppend(player_list);
            item.i = &meta;
            item.Init();
            player_hash[item.i->id()] = &item;
        }

        for (auto& meta : mapthing_meta_list) {
            MetaData::MapThing& item = a8::FastAppend(mapthing_list);
            item.i = &meta;
            mapthing_hash[item.i->thing_id()] = &item;
        }

        for (auto& meta : drop_meta_list) {
            MetaData::Drop& item = a8::FastAppend(drop_list);
            item.i = &meta;
            item.Init();
            drop_hash[item.i->drop_id()] = &item;
        }

        for (auto& meta : airdrop_meta_list) {
            MetaData::AirDrop& item = a8::FastAppend(airdrop_list);
            item.i = &meta;
            item.Init();
            auto itr = airdrop_hash.find(meta.map_id());
            if (itr != airdrop_hash.end()) {
                itr->second.push_back(&item);
            } else {
                airdrop_hash[meta.map_id()] = std::vector<MetaData::AirDrop*>({&item});
            }
        }

        {
            for (auto& meta : building_meta_list) {
                MetaData::Building& item = a8::FastAppend(building_list);
                item.i = &meta;
                item.Init();
                building_hash[item.i->mapid()] = &item;
            }
        }

        for (auto& meta : buff_meta_list) {
            if (!IsValidBuffEffect(meta.buff_effect())) {
                abort();
            }
            MetaData::Buff& item = a8::FastAppend(buff_list);
            item.i = &meta;
            item.Init();
            buff_hash[item.i->buff_id()] = &item;
        }

        for (auto& meta : skill_meta_list) {
            MetaData::Skill& item = a8::FastAppend(skill_list);
            item.i = &meta;
            item.Init();
            skill_hash[item.i->skill_id()] = &item;
        }

        for (auto& meta : attr_meta_list) {
            MetaData::Attr& item = a8::FastAppend(attr_list);
            item.i = &meta;
            attr_id_hash[item.i->attr_id()] = &item;
            attr_name_hash[item.i->attr_ename()] = &item;
        }

        for (auto& meta : rankreward_meta_list) {
            MetaData::RankReward& item = a8::FastAppend(rankreward_list);
            item.i = &meta;
            rankreward_hash[item.i->rank()] = &item;
        }

       for (auto& meta : killreward_meta_list) {
            MetaData::KillReward& item = a8::FastAppend(killreward_list);
            item.i = &meta;
            killreward_hash[item.i->kill_num()] = &item;
        }

       for (auto& meta : robot_meta_list) {
            MetaData::Robot& item = a8::FastAppend(robot_list);
            item.i = &meta;
            item.Init();
            robot_hash[meta.id()] = &item;
        }

    }

private:
    std::string res_path;
};

void MetaMgr::Init()
{
    loader_ = new MetaDataLoader();
    loader_->Load();
}

void MetaMgr::UnInit()
{
    delete loader_;
    loader_ = nullptr;
}

void MetaMgr::Reload()
{
    delete loader_;
    loader_ = nullptr;

    loader_ = new MetaDataLoader();
    loader_->Load();
}

int MetaMgr::GetSysParamAsInt(const std::string& param_name, int def_val)
{
    auto itr = loader_->parameter_hash.find(param_name);
    return itr != loader_->parameter_hash.end() ? itr->second->int_val: def_val;
}

double MetaMgr::GetSysParamAsFloat(const std::string& param_name, double def_val)
{
    auto itr = loader_->parameter_hash.find(param_name);
    return itr != loader_->parameter_hash.end() ? itr->second->float_val: def_val;
}

std::string MetaMgr::GetSysParamAsString(const std::string& param_name, const char* def_val)
{
    auto itr = loader_->parameter_hash.find(param_name);
    return itr != loader_->parameter_hash.end() ? itr->second->str_val: def_val;
}

MetaData::Map* MetaMgr::GetMap(int map_id)
{
    auto itr = loader_->gamemap_hash.find(map_id);
    return itr != loader_->gamemap_hash.end() ? itr->second : nullptr;
}

std::vector<MetaData::AirDrop*>* MetaMgr::GetMapAirDrop(int map_id)
{
    auto itr = loader_->airdrop_hash.find(map_id);
    return itr != loader_->airdrop_hash.end() ? &itr->second : nullptr;
}

MetaData::Map* MetaMgr::RandMap()
{
    std::vector<MetaData::Map*> map_list;
    for (auto& pair : loader_->gamemap_hash) {
        map_list.push_back(pair.second);
    }
    return !map_list.empty() ? map_list[rand() % map_list.size()] : nullptr;
}

MetaData::MapThing* MetaMgr::GetMapThing(int mapthing_id)
{
    auto itr = loader_->mapthing_hash.find(mapthing_id);
    return itr != loader_->mapthing_hash.end() ? itr->second : nullptr;
}

MetaData::Player* MetaMgr::GetPlayer(int id)
{
    auto itr = loader_->player_hash.find(id);
    return itr != loader_->player_hash.end() ? itr->second : nullptr;
}

MetaData::Equip* MetaMgr::GetEquip(int id)
{
    auto itr = loader_->equip_hash.find(id);
    return itr != loader_->equip_hash.end() ? itr->second : nullptr;
}

MetaData::Equip* MetaMgr::GetEquipBySlotId(int slot_id)
{
    auto itr = loader_->equip_slot_hash.find(slot_id);
    return itr != loader_->equip_slot_hash.end() ? itr->second : nullptr;
}

MetaData::Building* MetaMgr::GetBuilding(int building_id)
{
    auto itr = loader_->building_hash.find(building_id);
    return itr != loader_->building_hash.end() ? itr->second : nullptr;
}

MetaData::Drop* MetaMgr::GetDrop(int drop_id)
{
    auto itr = loader_->drop_hash.find(drop_id);
    return itr != loader_->drop_hash.end() ? itr->second : nullptr;
}

MetaData::SafeArea* MetaMgr::GetSafeArea(int area_id)
{
    auto itr = loader_->safearea_hash.find(area_id);
    return itr != loader_->safearea_hash.end() ? itr->second : nullptr;
}

std::vector<MetaData::MapTplThing>* MetaMgr::GetMapTplThing(const std::string& map_name)
{
    auto itr = loader_->maptpl_hash.find(map_name);
    return itr != loader_->maptpl_hash.end() ? &itr->second : nullptr;
}

std::vector<MetaData::MapTplThing>* MetaMgr::GetMapBornPoints(const std::string& map_name)
{
    auto itr = loader_->born_point_hash.find(map_name);
    return itr != loader_->born_point_hash.end() ? &itr->second : nullptr;
}

MetaData::MapTplThing* MetaMgr::GetMapAirDropPoint(const std::string& map_name, int point_id)
{
    auto itr = loader_->airdrop_point_hash.find(map_name);
    if (itr == loader_->airdrop_point_hash.end()) {
        return nullptr;
    }
    auto itr2 = itr->second.find(point_id);
    return itr2 != itr->second.end() ? &itr2->second : nullptr;
}

bool MetaMgr::GetMapTplSize(const std::string& map_name, int& width, int& height)
{
    auto itr = loader_->maptpl_size_hash.find(map_name);
    if (itr != loader_->maptpl_size_hash.end()) {
        width = std::get<0>(itr->second);
        height = std::get<1>(itr->second);
        return true;
    } else {
        return false;
    }
}

MetaData::Skill* MetaMgr::GetSkill(int skill_id)
{
    auto itr = loader_->skill_hash.find(skill_id);
    return itr != loader_->skill_hash.end() ? itr->second : nullptr;
}

MetaData::Buff* MetaMgr::GetBuff(int buff_id)
{
    auto itr = loader_->buff_hash.find(buff_id);
    return itr != loader_->buff_hash.end() ? itr->second : nullptr;
}

MetaData::Attr* MetaMgr::GetAttrById(int attr_id)
{
    auto itr = loader_->attr_id_hash.find(attr_id);
    return itr != loader_->attr_id_hash.end() ? itr->second : nullptr;
}

MetaData::Attr* MetaMgr::GetAttrByName(const std::string& attr_name)
{
    auto itr = loader_->attr_name_hash.find(attr_name);
    return itr != loader_->attr_name_hash.end() ? itr->second : nullptr;
}

MetaData::Driver* MetaMgr::GetDriver(int driver_id)
{
    auto itr = loader_->driver_hash.find(driver_id);
    return itr != loader_->driver_hash.end() ? itr->second : nullptr;
}

MetaData::Tank* MetaMgr::GetTank(int tank_id)
{
    auto itr = loader_->tank_hash.find(tank_id);
    return itr != loader_->tank_hash.end() ? itr->second : nullptr;
}

MetaData::TankSkin* MetaMgr::GetTankSkin(int tank_id)
{
    auto itr = loader_->tankskin_hash.find(tank_id);
    return itr != loader_->tankskin_hash.end() ? itr->second : nullptr;
}

MetaData::Cup* MetaMgr::GetCup(int rank)
{
    auto itr = loader_->cup_hash.find(rank);
    return itr != loader_->cup_hash.end() ? itr->second : nullptr;
}

int MetaMgr::RandTank()
{
    return !loader_->tankid_list.empty() ? loader_->tankid_list[rand() % loader_->tankid_list.size()] : 0;
}

float MetaMgr::GetRankRewardParam(int rank)
{
    auto itr = loader_->rankreward_hash.find(rank);
    return itr != loader_->rankreward_hash.end() ? itr->second->i->parameter() : 0.0f;
}

float MetaMgr::GetKillRewardParam(int kill_num)
{
    auto itr = loader_->killreward_hash.find(kill_num);
    return itr != loader_->killreward_hash.end() ? itr->second->i->parameter() : 0.0f;
}

std::vector<MetaData::Robot>* MetaMgr::GetRobotList()
{
    return &loader_->robot_list;
}

MetaData::Robot* MetaMgr::GetRobot(int robot_id)
{
    auto itr = loader_->robot_hash.find(robot_id);
    return itr != loader_->robot_hash.end() ? itr->second : nullptr;
}
