#include "precompile.h"

#include "playermgr.h"
#include "player.h"
#include "cs_proto.pb.h"
#include "ss_proto.pb.h"
#include "room.h"
#include "metamgr.h"

#include "framework/cpp/utils.h"

void PlayerMgr::Init()
{
}

void PlayerMgr::UnInit()
{
}

void PlayerMgr::_SS_WSP_SocketDisconnect(f8::MsgHdr& hdr, const ss::SS_WSP_SocketDisconnect& msg)
{
    Player* hum = GetPlayerBySocket(hdr.socket_handle);
    if (hum) {
        RemovePlayerBySocket(hdr.socket_handle);
        hum->room->OnPlayerOffline(hum);
    }
}

void PlayerMgr::_SS_Ping(f8::MsgHdr& hdr, const ss::SS_Ping& msg)
{
    ss::SS_Pong respmsg;
    GGListener::Instance()->SendToClient(hdr.socket_handle, 0, respmsg);
}

int PlayerMgr::OnlineNum()
{
    return socket_hash_.size();
}

Player* PlayerMgr::GetPlayerBySocket(int socket)
{
    auto itr = socket_hash_.find(socket);
    return itr != socket_hash_.end() ? itr->second : nullptr;
}

Player* PlayerMgr::CreatePlayerByCMJoin(long ip_saddr, int socket, const cs::CMJoin& msg)
{
    Player* hum = new Player();
    hum->socket_handle = socket;
    hum->ip_saddr = ip_saddr;
    hum->account_id = msg.account_id();
    hum->session_id = msg.session_id();
    hum->from_appid = msg.from_appid();
    hum->name = msg.name();
    hum->team_uuid = msg.team_uuid();
    hum->team_mode = msg.team_mode();
    hum->auto_fill = msg.auto_fill();
    hum->use_touch = msg.use_touch();
    hum->avatar_url = msg.avatar_url();
    hum->energy_shield = msg.energy_shield();
    hum->create_tick = a8::XGetTickCount();
    hum->account_registertime = f8::ExtractRegisterTimeFromSessionId(msg.session_id());
    for (auto& weapon : msg.weapons()) {
        if (weapon.weapon_id() != 0 && weapon.weapon_lv() > 0) {
            hum->weapon_configs[weapon.weapon_id()] = weapon.weapon_lv();
        }
    }
    for (auto& skin : msg.skins()) {
        if (skin.skin_id() != 0 && skin.skin_lv() > 0) {
            hum->tank_configs[skin.skin_id()] = skin.skin_lv();
        }
    }
    if (msg.weapon().weapon_id() != 0) {
        MetaData::Equip* weapon_meta = MetaMgr::Instance()->GetEquip(msg.weapon().weapon_id());
        if (weapon_meta) {
            hum->weapons[kGUN_SLOT0].weapon_idx = kGUN_SLOT0;
            hum->weapons[kGUN_SLOT0].weapon_id = msg.weapon().weapon_id();
            hum->weapons[kGUN_SLOT0].weapon_lv = msg.weapon().weapon_lv();
            hum->weapons[kGUN_SLOT0].meta = weapon_meta;
            #if 1
            hum->weapons[kGUN_SLOT0].ammo = hum->weapons[kGUN_SLOT0].GetClipVolume();
            #else
            hum->weapons[kGUN_SLOT0].ammo = kFIGHTING_MODE_BULLET_NUM;
            #endif
            hum->weapons[kGUN_SLOT0].Recalc();
            hum->curr_weapon = &hum->weapons[kGUN_SLOT0];
        }
    }
    hum->SetTankInfo(msg.baseskin());
    hum->driver.FromPB(&msg.driver());
    hum->tankskin.FromPB(&msg.tankskin());
    socket_hash_[socket] = hum;
    return hum;
}

void PlayerMgr::OnClientDisconnect(a8::XParams& param)
{
    int gg_socket = param.sender;

    std::vector<int> socket_list;
    for (auto& pair : socket_hash_) {
        unsigned short parent_socket_handle = (pair.first >> 16) & 0xFFFF;
        if (parent_socket_handle == gg_socket) {
            socket_list.push_back(pair.first);
        }
    }
    for (int socket_handle : socket_list) {
        Player* hum = GetPlayerBySocket(socket_handle);
        if (hum) {
            RemovePlayerBySocket(socket_handle);
            hum->room->OnPlayerOffline(hum);
        }
    }
}

void PlayerMgr::RemovePlayerBySocket(int socket_handle)
{
    auto itr = socket_hash_.find(socket_handle);
    if (itr != socket_hash_.end()) {
        itr->second->socket_handle = 0;
        socket_hash_.erase(itr);
    }
}
